@@ -3,6 +3,10 @@ from napalm.base.helpers import (
 )
 
 
+def rtt(key, iterable):
+    return (float(row.get(key, '-1ms').replace('ms', '')) for row in iterable)
+
+
 def to_seconds(time_format):
     seconds = minutes = hours = days = weeks = 0
 
@@ -22,7 +26,7 @@ def to_seconds(time_format):
         elif current_character == 'w':
             weeks = int(number_buffer)
         else:
-            raise ValueError('Invalid specifier - [{}]'.format(current_character))
+            raise ValueError(f'Invalid specifier - [{current_character}]')
         number_buffer = ''
 
     seconds += (minutes * 60)
